import json
import requests
import re
import sys
import time
from ansible.module_utils.basic import AnsibleModule, json, env_fallback
from collections import OrderedDict

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

def viptela_argument_spec():
    return dict(host=dict(type='str', required=True, fallback=(env_fallback, ['VMANAGE_HOST'])),
                port=dict(type='str', required=False, fallback=(env_fallback, ['VMANAGE_PORT'])),
                user=dict(type='str', required=True, fallback=(env_fallback, ['VMANAGE_USERNAME'])),
                password=dict(type='str', required=True, fallback=(env_fallback, ['VMANAGE_PASSWORD'])),
                validate_certs=dict(type='bool', required=False, default=False),
                timeout=dict(type='int', default=30)
                )


STANDARD_HTTP_TIMEOUT = 10
STANDARD_JSON_HEADER = {'Connection': 'keep-alive', 'Content-Type': 'application/json'}
POLICY_LIST_DICT = {
    'siteLists': 'site',
    'vpnLists': 'vpn',
}
VALID_STATUS_CODES = [200, 201, 202, 203, 204, 205, 206, 207, 208, 226]


class viptelaModule(object):

    def __init__(self, module, function=None):
        self.module = module
        self.params = module.params
        self.result = dict(changed=False)
        self.headers = dict()
        self.function = function
        self.cookies = None
        self.json = None

        self.method = None
        self.path = None
        self.response = None
        self.status = None
        self.url = None
        self.params['force_basic_auth'] = True
        self.user = self.params['user']
        self.password = self.params['password']
        self.host = self.params['host']
        self.timeout = self.params['timeout']
        self.modifiable_methods = ['POST', 'PUT', 'DELETE']

        self.session = requests.Session()
        self.session.verify = self.params['validate_certs']

        self.POLICY_DEFINITION_TYPES = ['cflowd', 'dnssecurity', 'control', 'hubandspoke', 'acl', 'vpnmembershipgroup',
                                        'mesh', 'rewriterule', 'data', 'rewriterule', 'aclv6']
        self.POLICY_LIST_TYPES = ['community', 'localdomain', 'ipv6prefix', 'dataipv6prefix', 'tloc', 'aspath', 'zone',
                                  'color', 'sla', 'app', 'mirror', 'dataprefix', 'extcommunity', 'site', 'ipprefixall',
                                  'prefix', 'umbrelladata', 'class', 'ipssignature', 'dataprefixall',
                                  'urlblacklist', 'policer', 'urlwhitelist', 'vpn']

        self.login()

    # Deleting (Calling destructor)
    # def __del__(self):
    #     self.logout()

    def _fallback(self, value, fallback):
        if value is None:
            return fallback
        return value

    def list_to_dict(self, list, key_name, remove_key=True):
        dict_value = OrderedDict()
        for item in list:
            if key_name in item:
                if remove_key:
                    key = item.pop(key_name)
                else:
                    key = item[key_name]

                dict_value[key] = item
            # else:
            #     self.fail_json(msg="key {0} not found in dictionary".format(key_name))

        return dict_value

    @staticmethod
    def compare_payloads(new_payload, old_payload, compare_values=None):
        if compare_values is None:
            compare_values = []
        payload_key_diff = []
        for key, value in new_payload.items():
            if key in compare_values:
                if key not in old_payload or new_payload[key] != old_payload[key]:
                    payload_key_diff.append(key)
        return payload_key_diff

    def login(self):
        # self.session.headers.update({'Connection': 'keep-alive', 'Content-Type': 'application/json'})

        try:
            response = self.session.post(
                url='https://{0}/j_security_check'.format(self.host),
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                data={'j_username': self.user, 'j_password': self.password},
                timeout=self.timeout
            )
        except requests.exceptions.RequestException as e:  # This is the correct syntax
            self.module.fail_json(msg=e)

        if response.text.startswith('<html>'):
            self.fail_json(msg='Could not login to device, check user credentials.', **self.result)

        response = self.session.get(
            url='https://{0}/dataservice/client/token'.format(self.host),
            timeout=self.timeout
        )
        if response.status_code == 200:
            self.session.headers['X-XSRF-TOKEN'] = response.content
        elif response.status_code == 404:
            # Assume this is pre-19.2
            pass
        else:
            self.fail_json(msg='Failed getting X-XSRF-TOKEN: {0}'.format(response.status_code))

        return response

    def logout(self):
        self.request('/dataservice/settings/clientSessionTimeout')
        self.request('/logout')

    def request(self, url_path, method='GET', data=None, files=None, headers=None, payload=None, status_codes=None):
        """Generic HTTP method for viptela requests."""

        if status_codes is None:
            status_codes = VALID_STATUS_CODES
        self.method = method
        self.url = 'https://{0}{1}'.format(self.host, url_path)
        self.result['url'] = self.url
        self.result['headers'] = self.session.headers.__dict__
        if files is None:
            self.session.headers['Content-Type'] = 'application/json'

        if payload:
            self.result['payload'] = payload
            data = json.dumps(payload)
            self.result['data'] = data

        response = self.session.request(method, self.url, files=files, data=data)

        self.status_code = response.status_code
        self.status = requests.status_codes._codes[response.status_code][0]
        decoded_response = {}
        if self.status_code not in status_codes:
            try:
                decoded_response = response.json()
            except JSONDecodeError:
                pass

            if 'error' in decoded_response:
                error = 'Unknown'
                details = 'Unknown'
                if 'details' in decoded_response['error']:
                    details = decoded_response['error']['details']
                if 'message' in decoded_response['error']:
                    error = decoded_response['error']['message']
                self.fail_json(msg='{0}: {1}'.format(error, details))
            else:
                self.fail_json(msg=self.status)

        try:
            response.json = response.json()
        except JSONDecodeError:
            response.json = {}

        return response

    def get_template_attachments(self, template_id, key='host-name'):
        response = self.request('/dataservice/template/device/config/attached/{0}'.format(template_id))

        attached_devices = []
        if response.json:
            device_list = response.json['data']
            for device in device_list:
                attached_devices.append(device[key])

        return attached_devices

    def generalTemplates_to_id(self, generalTemplates):
        converted_generalTemplates = []
        feature_templates = self.get_feature_template_dict(factory_default=True)
        for template in generalTemplates:
            if 'templateName' not in template:
                self.result['generalTemplates'] = generalTemplates
                self.fail_json(msg="Bad template")
            if template['templateName'] in feature_templates:
                template_item = {
                    'templateId': feature_templates[template['templateName']]['templateId'],
                    'templateType': template['templateType']}
                if 'subTemplates' in template:
                    subTemplates = []
                    for sub_template in template['subTemplates']:
                        if sub_template['templateName'] in feature_templates:
                            subTemplates.append(
                                {'templateId': feature_templates[sub_template['templateName']]['templateId'],
                                 'templateType': sub_template['templateType']})
                        else:
                            self.fail_json(msg="There is no existing feature template named {0}".format(
                                sub_template['templateName']))
                    template_item['subTemplates'] = subTemplates

                converted_generalTemplates.append(template_item)
            else:
                self.fail_json(msg="There is no existing feature template named {0}".format(template['templateName']))

        return converted_generalTemplates

    def convert_sequences_to_id(self, sequence_list):
        for sequence in sequence_list:
            for entry in sequence['match']['entries']:
                policy_list_dict = self.get_policy_list_dict(entry['listType'])
                if entry['listName'] in policy_list_dict:
                    entry['ref'] = policy_list_dict[entry['listName']]['listId']
                    entry.pop('listName')
                    entry.pop('listType')
                else:
                    self.fail_json(msg="Could not find list {0} of type {1}".format(entry['listName'], entry['listType']))
        return sequence_list

    def get_device_status(self, value, key='system-ip'):
        response = self.request('/dataservice/device?{0}={1}'.format(key, value))

        try:
            return response.json['data'][0]
        except:
            return {}

    def get_device_template_list(self, factory_default=False):
        response = self.request('/dataservice/template/device')

        return_list = []
        if response.json:

            device_body = response.json
            feature_template_dict = self.get_feature_template_dict(factory_default=True, key_name='templateId')

            for device in device_body['data']:
                object_response = self.request('/dataservice/template/device/object/{0}'.format(device['templateId']))
                if object_response.json:
                    object_value = object_response.json
                    if not factory_default and object_value['factoryDefault']:
                        continue

                    if 'generalTemplates' in object_value:
                        generalTemplates = []
                        for old_template in object_value.pop('generalTemplates'):
                            new_template = {
                                'templateName': feature_template_dict[old_template['templateId']]['templateName'],
                                'templateType': old_template['templateType']}
                            if 'subTemplates' in old_template:
                                subTemplates = []
                                for sub_template in old_template['subTemplates']:
                                    subTemplates.append({'templateName': feature_template_dict[sub_template['templateId']]['templateName'],
                                                         'templateType': sub_template['templateType']})
                                new_template['subTemplates'] = subTemplates

                            generalTemplates.append(new_template)
                        object_value['generalTemplates'] = generalTemplates

                    object_value['templateId'] = device['templateId']
                    object_value['attached_devices'] = self.get_template_attachments(device['templateId'])
                    object_value['input'] = self.get_template_input(device['templateId'])

                    return_list.append(object_value)

        return return_list

    def get_device_template_dict(self, factory_default=False, key_name='templateName', remove_key=True):
        device_template_list = self.get_device_template_list(factory_default=factory_default)

        return self.list_to_dict(device_template_list, key_name, remove_key)

    def get_feature_template_list(self, factory_default=False):
        response = self.request('/dataservice/template/feature')

        return_list = []
        if response.json:
            template_list = response.json['data']
            for template in template_list:
                if not factory_default and template['factoryDefault']:
                    continue
                template['templateDefinition'] = json.loads(template['templateDefinition'])
                template.pop('editedTemplateDefinition', None)
                return_list.append(template)

        return return_list

    def get_feature_template_dict(self, factory_default=False, key_name='templateName', remove_key=True):
        feature_template_list = self.get_feature_template_list(factory_default=factory_default)

        return self.list_to_dict(feature_template_list, key_name, remove_key)

    def get_policy_list(self, type, list_id):
        response = self.request('/dataservice/template/policy/list/{0}/{1}'.format(type.lower(), list_id))
        return response.json

    def get_policy_list_list(self, type='all'):
        if type == 'all':
            response = self.request('/dataservice/template/policy/list', status_codes=[200, 404])
        else:
            response = self.request('/dataservice/template/policy/list/{0}'.format(type.lower()), status_codes=[200, 404])

        if response.status_code == 404:
            return []
        else:
            return response.json['data']

    def get_policy_list_dict(self, type, key_name='name', remove_key=False):

        policy_list = self.get_policy_list_list(type)

        return self.list_to_dict(policy_list, key_name, remove_key=remove_key)

    def get_policy_definition(self, type, definition_id):
        response = self.request('/dataservice/template/policy/definition/{0}/{1}'.format(type, definition_id))
        return response.json

    def get_policy_definition_list(self, type):
        response = self.request('/dataservice/template/policy/definition/{0}'.format(type))

        return response.json['data']

    def get_vmanage_org(self):
        response = self.request('/dataservice/settings/configuration/organization')
        try:
            return response.json['data'][0]['org']
        except:
            return None

    def set_vmanage_org(self, org):
        payload = {'org': org}
        response = self.request('/dataservice/settings/configuration/organization', method='POST', payload=payload)

        return response.json['data']

    def get_vmanage_vbond(self):
        response = self.request('/dataservice/settings/configuration/device')
        try:
            return {'vbond': response.json['data'][0]['domainIp'], 'vbond_port': response.json['data'][0]['port']}
        except:
            return {'vbond': None, 'vbond_port': None}

    def set_vmanage_vbond(self, vbond, vbond_port='12346'):
        payload = {'domainIp': vbond, 'port': vbond_port}
        self.request('/dataservice/settings/configuration/device', method='POST', payload=payload)
        return

    def get_vmanage_ca_type(self):
        response = self.request('/dataservice/settings/configuration/certificate')
        try:
            return response.json['data'][0]['certificateSigning']
        except:
            return None

    def set_vmanage_ca_type(self, type):
        payload = {'certificateSigning': type, 'challengeAvailable': 'false'}
        self.request('/dataservice/settings/configuration/certificate', method='POST', payload=payload)
        return

    def get_vmanage_root_cert(self):
        response = self.request('/dataservice/certificate/rootcertificate')
        try:
            return response.json['rootcertificate']
        except:
            return None

    def set_vmanage_root_cert(self, cert):
        payload = {'enterpriseRootCA': cert}
        self.request('/dataservice/settings/configuration/certificate/enterpriserootca', method='PUT', payload=payload)
        return

    def install_device_cert(self, cert):
        response = self.request('/dataservice/certificate/install/signedCert', method='POST', data=cert)
        if response.json and 'id' in response.json:
            self.waitfor_action_completion(response.json['id'])
        else:
            self.fail_json(msg='Did not get action ID after attaching device to template.')
        return response.json['id']

    def get_policy_definition_dict(self, type, key_name='name', remove_key=False):

        policy_definition_list = self.get_policy_definition_list(type)

        return self.list_to_dict(policy_definition_list, key_name, remove_key=remove_key)

    def get_central_policy_list(self):
        response = self.request('/dataservice/template/policy/vsmart')
        if response.json:
            central_policy_list = response.json['data']
            for policy in central_policy_list:
                policy['policyDefinition'] = json.loads(policy['policyDefinition'])
                for item in policy['policyDefinition']['assembly']:
                    policy_definition = self.get_policy_definition(item['type'], item['definitionId'])
                    item['definitionName'] = policy_definition['name']
                    for entry in item['entries']:
                        for key, list in entry.items():
                            if key in POLICY_LIST_DICT:
                                for index, list_id in enumerate(list):
                                    policy_list = self.get_policy_list(POLICY_LIST_DICT[key], list_id)
                                    list[index] = policy_list['name']
            #     if 'policyDefinition' in policy:
            #         for old_template in policy.pop('policyDefinition'):
            #

            return central_policy_list
        else:
            return []

    def get_central_policy_dict(self, key_name='policyName', remove_key=False):

        central_policy_list = self.get_central_policy_list()

        return self.list_to_dict(central_policy_list, key_name, remove_key=remove_key)

    def get_unused_device(self, model):
        response = self.request('/dataservice/system/device/vedges?model={0}&state=tokengenerated'.format(model))

        if response.json:
            try:
                return response.json['data'][0]
            except:
                return response.json['data']
        else:
            return {}

    def get_device_by_state(self, state, type='vedges'):
        response = self.request('/dataservice/system/device/{0}?state={1}'.format(type, state))

        if response.json:
            try:
                return response.json['data'][0]
            except:
                return response.json['data']
        else:
            return {}

    def get_device_by_uuid(self, uuid, type='vedges'):
        response = self.request('/dataservice/system/device/{0}?uuid={1}'.format(type, uuid))

        if response.json:
            try:
                return response.json['data'][0]
            except:
                return response.json['data']
        else:
            return {}

    def get_device_by_device_ip(self, device_ip, type='vedges'):
        response = self.request('/dataservice/system/device/{0}?deviceIP={1}'.format(type, device_ip))

        if response.json:
            try:
                return response.json['data'][0]
            except:
                return response.json['data']
        else:
            return {}

    def get_device_by_name(self, name, type='vedges'):
        device_dict = self.get_device_dict(type)

        try:
            return device_dict[name]
        except:
            return {}

    def get_device_list(self, type, key_name='host-name', remove_key=True):
        response = self.request('/dataservice/system/device/{0}'.format(type))

        if response.json:
            return response.json['data']
        else:
            return []

    def get_device_dict(self, type, key_name='host-name', remove_key=False):

        device_list = self.get_device_list(type)

        return self.list_to_dict(device_list, key_name=key_name, remove_key=remove_key)

    def get_device_status_list(self):
        response = self.request('/dataservice/device')

        if response.json:
            return response.json['data']
        else:
            return []

    def get_device_status_dict(self, key_name='host-name', remove_key=False):

        device_list = self.get_device_list()

        return self.list_to_dict(device_list, key_name=key_name, remove_key=remove_key)

    def get_device_vedges(self, key_name='host-name', remove_key=True):
        response = self.request('/dataservice/system/device/vedges')

        if response.json:
            return self.list_to_dict(response.json['data'], key_name=key_name, remove_key=remove_key)
        else:
            return {}

    def get_device_controllers(self, key_name='host-name', remove_key=True):
        response = self.request('/dataservice/system/device/controllers')

        if response.json:
            return self.list_to_dict(response.json['data'], key_name=key_name, remove_key=remove_key)
        else:
            return {}

    def create_controller(self, device_ip, personality, username, password):
        payload = {
            "deviceIP": device_ip,
            "username": username,
            "password": password,
            "personality": personality,
            "generateCSR": "false"
        }

        response = self.request('/dataservice/system/device', method='POST', payload=payload)

        if response.json:
            return response.json
        else:
            return None

    def delete_controller(self, uuid):
        response = self.request('/dataservice/certificate/{0}'.format(uuid), method='DELETE')

        return response

    def decommision_device(self, uuid):
        response = self.request('/dataservice/system/device/decommission/{0}'.format(uuid), method='PUT')

        return response

    def generate_csr(self, device_ip):
        payload = {"deviceIP": device_ip}
        response = self.request('/dataservice/certificate/generate/csr', method='POST', payload=payload)

        if response.json:
            try:
                return response.json['data'][0]['deviceCSR']
            except:
                return None
        else:
            return None

    def generate_bootstrap(self, uuid, version='v1'):
        if version == 'v2':
            response = self.request('/dataservice/system/device/bootstrap/device/{0}?configtype=cloudinit&inclDefRootCert=true&version=v2'.format(uuid))
        else:
            response = self.request('/dataservice/system/device/bootstrap/device/{0}?configtype=cloudinit'.format(uuid))

        try:
            bootstrap_config = response.json['bootstrapConfig']
        except:
            return None

        regex = re.compile(r'otp : (?P<otp>[a-z0-9]+)[^a-z0-9]')
        match = regex.search(bootstrap_config)
        if match:
            otp = match.groups('otp')[0]
        else:
            otp = None

        return_dict = {
            'bootstrapConfig': bootstrap_config,
            'otp': otp,
            'uuid': uuid
        }
        return return_dict

    def get_template_input(self, template_id):
        payload = {
            "deviceIds": [],
            "isEdited": False,
            "isMasterEdited": False,
            "templateId": template_id
        }
        return_dict = {
            "columns": [],
        }
        response = self.request('/dataservice/template/device/config/input', method='POST', payload=payload)

        if response.json:
            if 'header' in response.json and 'columns' in response.json['header']:
                column_list = response.json['header']['columns']

                regex = re.compile(r'\((?P<variable>[^(]+)\)')

                for column in column_list:
                    if column['editable']:
                        match = regex.search(column['title'])
                        if match:
                            variable = match.groups('variable')[0]
                        else:
                            variable = None

                        entry = {'title': column['title'],
                                 'property': column['property'],
                                 'variable': variable}
                        return_dict['columns'].append(entry)

        return return_dict

    def get_template_variables(self, template_id):
        payload = {
            "deviceIds": [],
            "isEdited": False,
            "isMasterEdited": False,
            "templateId": template_id
        }
        return_dict = {}
        response = self.request('/dataservice/template/device/config/input', method='POST', payload=payload)

        if response.json:
            if 'header' in response.json and 'columns' in response.json['header']:
                column_list = response.json['header']['columns']

                regex = re.compile(r'\((?P<variable>[^(]+)\)')

                for column in column_list:
                    if column['editable']:
                        # match = regex.search(column['title'])
                        match = regex.findall(column['title'])
                        if match:
                            # variable = match.groups('variable')[0]
                            variable = match[-1]
                            return_dict[variable] = column['property']

        return return_dict

    def get_template_optional_variables(self, template_id):
        payload = {
            "deviceIds": [],
            "isEdited": False,
            "isMasterEdited": False,
            "templateId": template_id
        }
        return_dict = {}
        response = self.request('/dataservice/template/device/config/input', method='POST', payload=payload)

        if response.json:
            if 'header' in response.json and 'columns' in response.json['header']:
                column_list = response.json['header']['columns']

                regex = re.compile(r'\((?P<variable>[^(]+)\)')

                # The following can be removed once the API will mark as optional
                # the nexthop value of a static route that has been marked as optional
                regexYangStaticR = re.compile(r'.*/vpn-instance/ip/route/.*/prefix')
                regex_yang_nexthop = re.compile(r'.*/vpn-instance/ip/route/(?P<staticR>.*)/next-hop/.*/address')
                optionalStaticRoutesList = []
                # Until here

                # The following can be removed once the API will mark as optional
                # all the vrrp attributes once the vrrp grp-id has been marked optional
                regexYangVRRPgrpID = re.compile(r'.*/vrrp/.*/grp-id')
                regexYangVRRPpriority = re.compile(r'.*/vrrp/(?P<VVRPgrp>.*)/priority')
                regexYangVRRPtimer = re.compile(r'.*/vrrp/(?P<VVRPgrp>.*)/timer')
                regexYangVRRPtrackPrefix = re.compile(r'.*/vrrp/(?P<VVRPgrp>.*)/track-prefix-list')
                regexYangVRRPtrackOMP = re.compile(r'.*/vrrp/(?P<VVRPgrp>.*)/track-omp')
                regexYangVRRPipAddress = re.compile(r'.*/vrrp/(?P<VVRPgrp>.*)/ipv4/address')
                optionalVRRPvariales = []
                # Until here

                # The following can be removed once the API will mark as optional
                # all the logging attributes once the logging has been marked optional
                regexYangLoggingServerName = re.compile(r'///logging/server/.*/name')
                regexYangLoggingSourceInt = re.compile(r'///logging/server/(?P<LoggServer>.*)/source-interface')
                regexYangLoggingVPN = re.compile(r'///logging/server/(?P<LoggServer>.*)/vpn')
                regexYangLoggingPriority = re.compile(r'///logging/server/(?P<LoggServer>.*)/priority')
                optionalLoggingVariales = []
                # Until here

                for column in column_list:

                    # The following can be removed once the API will mark as optional
                    # the nexthop value of a static route that has been marked as optional

                    # Based on the regular expression above we match
                    # static routes and next-hop variables based on the YANG variable
                    # a static route looks like this /1/vpn-instance/ip/route/<COMMON_NAME_OF_THE_ROUTE>/prefix
                    # a next-hop looks like this /1/vpn-instance/ip/route/<COMMON_NAME_OF_THE_ROUTE>/next-hop/<COMMON_NAME_OF_THE_NH>/address

                    # If we find a static route and this is optional we
                    # store its common name into an array
                    # we don't add this parameter to the return list now
                    # since it will be added later
                    isStaticR = regexYangStaticR.match(column['property'])
                    if isStaticR and column['optional']:
                        match = regex.findall(column['title'])
                        if match:
                            variable = match[-1]
                            optionalStaticRoutesList.append(variable)

                    # If we find a next-hop we extrapolate the common name
                    # of the static route. If we have already found that
                    # common name and we know it is optional we will add
                    # this next-hop paramter to the return list since it
                    # will be optional as well

                    # ALL OF THIS IS BASED ON THE ASSUMPTION THAT STATIC ROUTES
                    # ARE LISTED BEFORE NEXT-HOP VALUES
                    nextHopStaticR = regex_yang_nexthop.findall(column['property'])
                    if nextHopStaticR:
                        if nextHopStaticR[0] in optionalStaticRoutesList:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']

                    # Until here

                    # The following can be removed once the API will mark as optional
                    # the attributes for vrrp as optional if the whole vrrp has been
                    # marked as optional

                    # Based on the regular expression above we match
                    # vrrp atributes based on the YANG variable

                    # If we find a VRRP grp ID and this is optional we
                    # store its common name into an array
                    # we don't add this parameter to the return list now
                    # since it will be added later
                    isVRRP = regexYangVRRPgrpID.match(column['property'])
                    if isVRRP and column['optional']:
                        match = regex.findall(column['title'])
                        if match:
                            variable = match[-1]
                            optionalVRRPvariales.append(variable)

                    # If we find a any vrrp attribute we extrapolate the common name
                    # If we have already found that
                    # common name and we know it is optional we will add
                    # this paramter to the return list since it
                    # will be optional as well

                    # ALL OF THIS IS BASED ON THE ASSUMPTION THAT VRRP GRP-ID is
                    # LISTED BEFORE ALL THE OTHER ATTRIBUTES
                    VRRPpriority = regexYangVRRPpriority.findall(column['property'])
                    VRRPtimer = regexYangVRRPtimer.findall(column['property'])
                    VRRPtrackPrefix = regexYangVRRPtrackPrefix.findall(column['property'])
                    VRRPipAddress = regexYangVRRPipAddress.findall(column['property'])
                    VRRPtrackOMP = regexYangVRRPtrackOMP.findall(column['property'])
                    if VRRPpriority:
                        if VRRPpriority[0] in optionalVRRPvariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif VRRPtimer:
                        if VRRPtimer[0] in optionalVRRPvariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif VRRPtrackPrefix:
                        if VRRPtrackPrefix[0] in optionalVRRPvariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif VRRPipAddress:
                        if VRRPipAddress[0] in optionalVRRPvariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif VRRPtrackOMP:
                        if VRRPtrackOMP[0] in optionalVRRPvariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                                # Until here

                    # Same logic for logging optional variables
                    isLogging = regexYangLoggingServerName.match(column['property'])
                    if isLogging and column['optional']:
                        match = regex.findall(column['title'])
                        if match:
                            variable = match[-1]
                            optionalLoggingVariales.append(variable)

                    LoggingSourceInt = regexYangLoggingSourceInt.findall(column['property'])
                    LoggingVPN = regexYangLoggingVPN.findall(column['property'])
                    LoggingPriority = regexYangLoggingPriority.findall(column['property'])

                    if LoggingSourceInt:
                        if LoggingSourceInt[0] in optionalLoggingVariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif LoggingVPN:
                        if LoggingVPN[0] in optionalLoggingVariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']
                    elif LoggingPriority:
                        if LoggingPriority[0] in optionalLoggingVariales:
                            match = regex.findall(column['title'])
                            if match:
                                variable = match[-1]
                                return_dict[variable] = column['property']

                    # Until here

                    if column['editable'] and column['optional']:
                        match = regex.findall(column['title'])
                        if match:
                            variable = match[-1]
                            return_dict[variable] = column['property']

        return return_dict

    def get_software_images_list(self):
        # TODO undertand the difference with the URL: /dataservice/device/action/software/images used in devnetsandbox
        response = self.request('/dataservice/device/action/software', method='GET')

        if response.json:
            return response.json['data']
        else:
            return []

    def get_installed_software(self, type):
        response = self.request('/dataservice/device/action/install/devices/{0}?groupId=all'.format(type), method='GET')

        if response.json:
            return response.json['data']
        else:
            return []

    def software_install(self, devices, deviceType, data, reboot):

        payload = {
            "action": "install",
            "input": {
                "vEdgeVPN": 0,
                "vSmartVPN": 0,
                "data": data,
                "versionType": "vmanage",
                "reboot": reboot,
                "sync": True
            },
            "devices": devices,
            "deviceType": deviceType
        }

        response = self.request('/dataservice/device/action/install', method='POST', payload=payload)

        if response.json and 'id' in response.json:
            self.waitfor_action_completion(response.json['id'])
        else:
            self.fail_json(
                msg='Did not get action ID after installing software.')

        return response.json['id']

    def set_default_partition(self, devices, deviceType):

        payload = {
            "action": "defaultpartition",
            "devices": devices,
            "deviceType": deviceType
        }

        response = self.request('/dataservice/device/action/defaultpartition', method='POST', payload=payload)

        if response.json and 'id' in response.json:
            self.waitfor_action_completion(response.json['id'])
        else:
            self.fail_json(
                msg='Did not get action ID after setting default image.')

        return response.json['id']

    def push_certificates(self):
        response = self.request('/dataservice/certificate/vedge/list?action=push', method='POST')
        if response.json and 'id' in response.json:
            self.waitfor_action_completion(response.json['id'])
        else:
            self.fail_json(msg='Did not get action ID after pushing certificates.')
        return response.json['id']

    def reattach_device_template(self, template_id, process_id=None):
        device_list = self.get_template_attachments(template_id, key='uuid')
        # First, we need to get the input to feed to the re-attach
        payload = {
            "templateId": template_id,
            "deviceIds": device_list,
            "isEdited": "true",
            "isMasterEdited": "false"
        }
        response = self.request('/dataservice/template/device/config/input/', method='POST', payload=payload)
        # Then we feed that to the attach
        if response.json and 'data' in response.json:
            payload = {
                "deviceTemplateList":
                    [
                        {
                            "templateId": template_id,
                            "device": response.json['data'],
                            "isEdited": "true"
                        }
                    ]
            }
            response = self.request('/dataservice/template/device/config/attachfeature', method='POST', payload=payload)
            if response.json and 'id' in response.json:
                self.waitfor_action_completion(response.json['id'])
            else:
                self.fail_json(
                    msg='Did not get action ID after attaching device to template.')

            # if process_id:
            #    self.request('/dataservice/template/lock/{0}'.format(process_id), method='DELETE')

        else:
            self.fail_json(msg="Could not retrieve input for template {0}".format(template_id))
        return response.json['id']

    def waitfor_action_completion(self, action_id):
        status = 'in_progress'
        response = {}
        while status == "in_progress":
            response = self.request('/dataservice/device/action/status/{0}'.format(action_id))
            if response.json:
                status = response.json['summary']['status']
                if 'data' in response.json and response.json['data']:
                    action_status = response.json['data'][0]['statusId']
                    action_activity = response.json['data'][0]['activity']
                    if 'actionConfig' in response.json['data'][0]:
                        action_config = response.json['data'][0]['actionConfig']
                    else:
                        action_config = None
            else:
                self.fail_json(msg="Unable to get action status: No response")
            time.sleep(10)

        # self.result['action_response'] = response.json
        self.result['action_id'] = action_id
        self.result['action_status'] = action_status
        self.result['action_activity'] = action_activity
        self.result['action_config'] = action_config
        if self.result['action_status'] == 'failure':
            self.fail_json(msg="Action failed")
        return response

    def exit_json(self, **kwargs):
        # self.logout()
        """Custom written method to exit from module."""

        self.result.update(**kwargs)
        self.module.exit_json(**self.result)

    def fail_json(self, msg, **kwargs):
        # self.logout()
        """Custom written method to return info on failure."""

        self.result.update(**kwargs)
        self.module.fail_json(msg=msg, **self.result)



"""Cisco vManage Authentication API Methods.
"""

from __future__ import (absolute_import, division, print_function)

import requests
import urllib3
from vmanage.api.utilities import Utilities

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Authentication(object):
    """vManage Authentication API

    Responsible for retrieving the JSESSIONID after a username/password
    has been authenticated.  If the vManage version is >= 19.2.0 then
    the X-XSRF-TOKEN will be retrieved and added to the header.  An
    HTTP(S) Request session object will be returned.

    """
    def __init__(self, host=None, user=None, password=None, port=443, validate_certs=False, timeout=10):
        """Initialize Authentication object with session parameters.

        Args:
            host (str): hostname or IP address of vManage
            user (str): username for authentication
            password (str): password for authentication
            port (int): default HTTPS port 443
            validate_certs (bool): turn certificate validation
                on or off.
            timeout (int): how long Reqeusts will wait for a
                response from the server, default 10 seconds

        """

        self.host = host
        self.user = user
        self.password = password
        self.port = port
        self.timeout = timeout
        self.base_url = f'https://{self.host}:{self.port}/dataservice/'
        self.session = requests.Session()
        self.session.verify = validate_certs

    def login(self):
        """Executes login tasks against vManage to retrieve token(s).

        Args:
            None.

        Returns:
            self.session: a Requests session with JSESSIONID and an
            X-XSRF-TOKEN for vManage version >= 19.2.0.

        Raises:
            LoginFailure: If the username/password are incorrect.
            RequestException: If the host is not accessible.

        """

        try:
            api = 'j_security_check'
            url = f'{self.base_url}{api}'
            response = self.session.post(url=url,
                                         data={
                                             'j_username': self.user,
                                             'j_password': self.password
                                         },
                                         timeout=self.timeout)

            if (response.status_code != 200 or response.text.startswith('<html>')):
                raise ConnectionError('Login failed, check user credentials.')

            version = Utilities(self.session, self.host, self.port).get_vmanage_version()

            if version >= '19.2.0':
                api = 'client/token'
                url = f'{self.base_url}{api}'
                response = self.session.get(url=url, timeout=self.timeout)
                self.session.headers['X-XSRF-TOKEN'] = response.content

        except requests.exceptions.RequestException as e:
            raise ConnectionError(f'Could not connect to {self.host}: {e}')

        return self.session



import click
from vmanage.cli.certificate.push import push
from vmanage.cli.certificate.generate_csr import generate_csr
from vmanage.cli.certificate.install import install


@click.group()
def certificate():
    """
    Certficate commands
    """


certificate.add_command(push)
certificate.add_command(generate_csr)
certificate.add_command(install)



import click
from vmanage.cli.activate import activate
from vmanage.cli.deactivate import deactivate
from vmanage.cli.decommission import decommission
from vmanage.cli.show import show
from vmanage.cli.export import export
from vmanage.cli.import_cmd import import_cmd
from vmanage.cli.clean import clean
from vmanage.cli.certificate import certificate
from vmanage.cli.set_cmd import set_cmd
from vmanage.cli.reset import reset
from vmanage.api.authentication import Authentication

# from vmanage.api.big import vmanage_session


class CatchAllExceptions(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.vmanage(*args, **kwargs)
        except Exception as exc:
            click.secho('Exception raised while running your command', fg="red")
            click.secho("Please open an issue and provide this info:", fg="red")
            click.secho("%s" % exc, fg="red")


class Viptela(object):
    def __init__(self, host, port, username, password):
        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.__auth = None

    # use this to defer authentication until it's needed
    @property
    def auth(self):
        if self.__auth is None:
            self.__auth = Authentication(host=self.host, port=self.port, user=self.username,
                                         password=self.password).login()
        return self.__auth


# @click.group(cls=CatchAllExceptions)
@click.group()
@click.option('--host', envvar='VMANAGE_HOST', help='vManage Host (env: VMANAGE_HOST)', required=True)
@click.option('--port',
              envvar='VMANAGE_PORT',
              help='vManage Port (env: VMANAGE_PORT, default=443)',
              default=443,
              required=False)
@click.option('--username', envvar='VMANAGE_USERNAME', help='vManage Username (env: VMANAGE_USERNAME)', required=True)
@click.option('--password',
              envvar='VMANAGE_PASSWORD',
              prompt=True,
              hide_input=True,
              help='vManage Password (env: VMANAGE_PASSWORD)',
              required=True)
@click.pass_context
def vmanage(ctx, host, port, username, password):
    ctx.obj = Viptela(host, port, username, password)


vmanage.add_command(activate)
vmanage.add_command(deactivate)
vmanage.add_command(decommission)
vmanage.add_command(show)
vmanage.add_command(export)
vmanage.add_command(import_cmd)
vmanage.add_command(certificate)
vmanage.add_command(clean)
vmanage.add_command(set_cmd)
vmanage.add_command(reset)



"""Cisco vManage Certificate API Methods.
"""

import json
from vmanage.api.http_methods import HttpMethods
from vmanage.data.parse_methods import ParseMethods
from vmanage.api.utilities import Utilities


class Certificate(object):
    """vManage Certificate API

    Responsible for DELETE, GET, POST, PUT methods against vManage
    Certificates.

    """
    def __init__(self, session, host, port=443):
        """Initialize Device Inventory object with session parameters.

        Args:
            session (obj): Requests Session object
            host (str): hostname or IP address of vManage
            port (int): default HTTPS 443

        """

        self.session = session
        self.host = host
        self.port = port
        self.base_url = f'https://{self.host}:{self.port}/dataservice/certificate/'

    def generate_csr(self, device_ip):
        """Generate CSR for device

        Args:
            device_ip (str): IP address of device.

        Returns:
            deviceCSR (str): The CSR for the device.
        """

        payload = {"deviceIP": device_ip}
        url = f"{self.base_url}generate/csr"
        response = HttpMethods(self.session, url).request('POST', payload=json.dumps(payload))
        result = ParseMethods.parse_data(response)
        return result[0]['deviceCSR']

    def install_device_cert(self, cert):
        """Install signed cert on vManage

        Args:
            cert (str): The certificate to install.

        Returns:
            id (str): The action ID of the install command.
        """

        url = f"{self.base_url}install/signedCert"
        response = HttpMethods(self.session, url).request('POST', payload=cert)
        utilities = Utilities(self.session, self.host)
        action_id = ParseMethods.parse_id(response)
        utilities.waitfor_action_completion(action_id)

        return action_id

    def push_certificates(self):
        """Push certificates to all controllers

        Returns:
            id (str): The action ID of the push command.
        """

        url = f"{self.base_url}vedge/list?action=push"
        response = HttpMethods(self.session, url).request('POST', payload={})
        utilities = Utilities(self.session, self.host)
        action_id = ParseMethods.parse_id(response)
        utilities.waitfor_action_completion(action_id)

        return action_id

    def get_vmanage_root_cert(self):
        """Get vManage root certificate

        Args:

        Returns:
            rootcertificate (str): The root certificate.
        """

        url = f"{self.base_url}rootcertificate"
        response = HttpMethods(self.session, url).request('GET')
        try:
            return response['json']['rootcertificate']
        except:
            raise Exception('Error retrieving root certificate')



import click
from vmanage.api.settings import Settings


@click.command('ca-type')
@click.pass_obj
def ca_type(ctx):
    """
    Get vManage CA type
    """

    vmanage_settings = Settings(ctx.auth, ctx.host, ctx.port)
    result = vmanage_settings.get_vmanage_ca_type()
    click.echo(result)



import click
from vmanage.cli.import_cmd.templates import templates
from vmanage.cli.import_cmd.policies import policies
from vmanage.cli.import_cmd.attachments import attachments
from vmanage.cli.import_cmd.serial_file import serial_file
from vmanage.cli.import_cmd.root_cert import root_cert


@click.group('import')
def import_cmd():
    """
    Import commands
    """


import_cmd.add_command(templates)
import_cmd.add_command(policies)
import_cmd.add_command(attachments)
import_cmd.add_command(serial_file)
import_cmd.add_command(root_cert)



import click
from vmanage.cli.export.templates import templates
from vmanage.cli.export.policies import policies
from vmanage.cli.export.attachments import attachments


@click.group()
def export():
    """
    Export commands
    """


export.add_command(templates)
export.add_command(policies)
export.add_command(attachments)



#!/usr/bin/env python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

from ansible.module_utils.basic import AnsibleModule, json
from ansible.module_utils.viptela.vmanage import Vmanage, vmanage_argument_spec
from vmanage.data.policy_data import PolicyData


def run_module():
    # define available arguments/parameters a user can pass to the module
    argument_spec = vmanage_argument_spec()
    argument_spec.update(type=dict(type='str', required=False, default='all'),
                         )

    # seed the result dict in the object
    # we primarily care about changed and state
    # change is if this module effectively modified the target
    # state will include any data that you want your module to pass back
    # for consumption, for example, in a subsequent task
    result = dict(
        changed=False,
        original_message='',
        message=''
    )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    vmanage = Vmanage(module)
    policy_data = PolicyData(vmanage.auth, vmanage.host)

    vmanage.result['policy_definitions'] = policy_data.export_policy_definition_list(vmanage.params['type'])

    vmanage.exit_json(**vmanage.result)


def main():
    run_module()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

from ansible.module_utils.basic import AnsibleModule, json
from ansible.module_utils.viptela.viptela import viptelaModule, viptela_argument_spec
from collections import OrderedDict


def run_module():
    # define available arguments/parameters a user can pass to the module
    argument_spec = viptela_argument_spec()
    argument_spec.update(state=dict(type='str', choices=['absent', 'present'], default='present'),
                         name=dict(type='str'),
                         transport_ip=dict(type='str', aliases=['device_ip', 'system_ip']),
                         uuid=dict(type='str', alias='deviceIP'),
                         personality=dict(type='str', choices=['vmanage', 'vsmart', 'vbond', 'vedge'], default='vedge'),
                         device_username=dict(type='str', alias='device_user'),
                         device_password=dict(type='str')
                         )

    # seed the result dict in the object
    # we primarily care about changed and state
    # change is if this module effectively modified the target
    # state will include any data that you want your module to pass back
    # for consumption, for example, in a subsequent task
    result = dict(
        changed=False,
    )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    viptela = viptelaModule(module)
    viptela.result['what_changed'] = []

    if viptela.params['personality'] == 'vedge':
        device_type = 'vedges'
    else:
        device_type = 'controllers'

    device = {}
    if viptela.params['uuid']:
        device = viptela.get_device_status(viptela.params['uuid'], key='uuid')
    # See if we can find the device by deviceIP
    if not device and viptela.params['transport_ip']:
        device = viptela.get_device_status(viptela.params['transport_ip'])
    # If we could not find the device by deviceIP, see if we can find it be (host)name
    if not device and viptela.params['name']:
        device = viptela.get_device_status(viptela.params['name'], key='host-name')

    viptela.result['device'] = device
    if viptela.params['state'] == 'present':
        if device:
            # Can't really change anything
            pass
        else:
            viptela.result['what_changed'].append('new')
            if not viptela.params['device_username'] or not viptela.params['device_password']:
                viptela.fail_json(msg="device_username and device_password must be specified when add a new device")
            if not module.check_mode:
                response = viptela.create_controller(viptela.params['transport_ip'], viptela.params['personality'],
                                                     viptela.params['device_username'], viptela.params['device_password'])
                viptela.result['response'] = response
    else:
        if device and device_type == 'controllers':
            if not module.check_mode:
                viptela.delete_controller(device['uuid'])
        elif device and device_type == 'vedges':
            device_config = viptela.get_device_by_uuid(device['uuid'], type=device_type)
            if 'vedgeCertificateState' in device_config and device_config['vedgeCertificateState'] != 'tokengenerated':
                viptela.result['what_changed'].append('delete')
                if not module.check_mode:
                    viptela.decommision_device(device['uuid'])

    if viptela.result['what_changed']:
        viptela.result['changed'] = True

    viptela.exit_json(**viptela.result)


def main():
    run_module()


if __name__ == '__main__':
    main()



"""Cisco vManage Centralized Policy API Methods.
"""
import json

from vmanage.api.http_methods import HttpMethods
from vmanage.api.policy_definitions import PolicyDefinitions
from vmanage.data.parse_methods import ParseMethods
from vmanage.utils import list_to_dict


class CentralPolicy(object):
    """vManage Central Policy API

    Responsible for DELETE, GET, POST, PUT methods against vManage
    Central Policy.

    """
    def __init__(self, session, host, port=443):
        """Initialize Centralized Policy object with session parameters.

        Args:
            session (obj): Requests Session object
            host (str): hostname or IP address of vManage
            port (int): default HTTPS 443

        """

        self.session = session
        self.host = host
        self.port = port
        self.base_url = f'https://{self.host}:{self.port}/dataservice/'
        self.policy_definitions = PolicyDefinitions(self.session, self.host)

    def activate_central_policy(self, policy_name, policy_id):
        """Activates the current active centralized policy

        Args:
            policyId (str): ID of the active centralized policy

        Returns:
            action_id (str): The Action ID from the activation.

        """

        url = f"{self.base_url}template/policy/vsmart/activate/{policy_id}?confirm=true"
        payload = {'policyName': policy_name}
        response = HttpMethods(self.session, url).request('POST', payload=json.dumps(payload))
        ParseMethods.parse_status(response)
        action_id = ParseMethods.parse_id(response)

        return action_id

    def reactivate_central_policy(self, policy_id):
        """reActivates the current active centralized policy

        Args:
            policyId (str): ID of the active centralized policy

        Returns:
            action_id (str): The Action ID from the activation.

        """

        url = f"{self.base_url}template/policy/vsmart/activate/{policy_id}?confirm=true"
        payload = {'isEdited': True}
        response = HttpMethods(self.session, url).request('POST', payload=json.dumps(payload))
        ParseMethods.parse_status(response)
        action_id = ParseMethods.parse_id(response)

        return action_id

    def deactivate_central_policy(self, policy_id):
        """Deactivates the current active centralized policy

        Args:
            policyId (str): ID of the deactive centralized policy

        Returns:
            result (str): The Action ID from the activation.

        """

        url = f"{self.base_url}template/policy/vsmart/deactivate/{policy_id}?confirm=true"
        response = HttpMethods(self.session, url).request('POST')
        ParseMethods.parse_status(response)
        if 'json' in response and 'id' in response['json']:
            return response['json']['id']

        return None

    def add_central_policy(self, policy):
        """Delete a Central Policy from vManage.

        Args:
            policy: The Central Policy

        Returns:
            result (dict): All data associated with a response.

        """
        url = f"{self.base_url}template/policy/vsmart"
        HttpMethods(self.session, url).request('POST', payload=json.dumps(policy))

    def update_central_policy(self, policy, policy_id):
        """Update a Central from vManage.

        Args:
            policy: The Central Policy
            policy_id: The ID of the Central Policy to update

        Returns:
            result (dict): All data associated with a response.

        """

        url = f"{self.base_url}template/policy/vsmart/{policy_id}"
        response = HttpMethods(self.session, url).request('PUT', payload=json.dumps(policy))
        return response

    def delete_central_policy(self, policyId):
        """Deletes the specified centralized policy

        Args:
            policyId (str): ID of the active centralized policy
        Returns:
            result (dict): All data associated with a response.

        """

        api = f"template/policy/vsmart/{policyId}"
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('DELETE')
        result = ParseMethods.parse_status(response)
        return result

    def get_central_policy(self):
        """Obtain a list of all configured central policies

        Returns:
            result (dict): All data associated with a response.

        """

        url = f"{self.base_url}template/policy/vsmart"
        response = HttpMethods(self.session, url).request('GET')
        result = ParseMethods.parse_data(response)
        return result

    def get_central_policy_list(self):
        """Get all Central Policies from vManage.

        Returns:
            response (list): A list of all policy lists currently
                in vManage.

        """

        central_policy_list = self.get_central_policy()
        # We need to convert the policy definitions from JSON
        for policy in central_policy_list:
            try:
                json_policy = json.loads(policy['policyDefinition'])
                policy['policyDefinition'] = json_policy
            except Exception:  # TODO: figuring out better exception type to catch
                pass
        return central_policy_list

    def get_central_policy_dict(self, key_name='policyName', remove_key=False):
        """Get all Central Policies from vManage.

        Args:
            key_name (str): The name of the attribute to use as the key
            remove_key (bool): Remove the key from the dict (default: False)

        Returns:
            response (dict): A dict of all Central Policies currently
                in vManage.

        """

        central_policy_list = self.get_central_policy_list()

        return list_to_dict(central_policy_list, key_name, remove_key=remove_key)



"""Cisco vManage Policy Lists API Methods.
"""

import json
import time
import sys
from vmanage.api.http_methods import HttpMethods
from vmanage.data.parse_methods import ParseMethods


class PolicyUpdates(object):
    """vManage Policy Updates API

    Responsible for POST and PUT methods against vManage
    Policy APIs used in Centralized, Localized, and Security Policy.

    """
    def __init__(self, session, host, port=443):
        """Initialize Policy Updates object with session parameters.

        Args:
            session (obj): Requests Session object
            host (str): hostname or IP address of vManage
            port (int): default HTTPS 443

        """

        self.session = session
        self.host = host
        self.port = port
        self.base_url = f'https://{self.host}:{self.port}/dataservice/'

    def get_device_ids(self, template_id):
        """GET vSmart device ids from vManage.

        Args:
            templateid (str): vManaged assigned template identifier.

        Returns:
            response (list): device ids list.

        """

        api = "template/device/config/attached/" + template_id
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('GET')
        result = ParseMethods.parse_data(response)
        device_ids = []
        for device in result:
            device_ids.append(device['uuid'])

        return device_ids

    def get_device_inputs(self, template_id, device_ids):
        """GET vSmart device inputs from vManage.

        Args:
            templateid (str): vManage assigned template identifier.
            device_ids (list): vManage assigned device ids for vSmarts.

        Returns:
            response (dict): device inputs.

        """

        payload = {"templateId": template_id, "deviceIds": device_ids, "isEdited": True, "isMasterEdited": False}

        api = "template/device/config/input"
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('POST', payload=json.dumps(payload))
        result = ParseMethods.parse_data(response)

        for item in result:
            item['csv-templateId'] = template_id

        return result

    def get_policy_id(self, policy_type, policy_name):
        """GET vSmart device ids from vManage.

        Args:
            policy_type (str): vManage policy type i.e. approute, data, hubandspoke etc.
            policy_name (str): policy name for user needs policy id.

        Returns:
            response (str): vManage assigned policy id.

        """

        api = "template/policy/definition/" + policy_type
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('GET')
        result = ParseMethods.parse_data(response)
        policy_id = ""

        for item in result:
            if item["name"] == policy_name:
                policy_id = item["definitionId"]
                break

        return policy_id

    def get_policy_definition(self, policy_type, policy_id):
        """GET vSmart device ids from vManage.

        Args:
            policy_type (str): vManage policy type i.e. approute, data, hubandspoke etc.
            policy_id (str): vManage assigned policy id.

        Returns:
            response (dict): policy definition.

        """

        api = "template/policy/definition/%s/%s" % (policy_type, policy_id)
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('GET')

        return response['json']

    def update_policy_definition(self, policy_type, name, policy_id, policy_def, new_color, seq_name=None):
        """update policy definition

        Args:
            policy_type (str): vManage policy type i.e. approute, data, hubandspoke etc.
            policy_id (str): vManage assigned policy id.
            policy_def (dict): dict of data for policy updates to be done

        Returns:
            response (dict): policy definition.

        """
        #pylint: disable=too-many-nested-blocks
        for item1 in policy_def["sequences"]:
            if item1["sequenceName"] == seq_name or seq_name is None:
                for item2 in item1["actions"]:
                    if item2['type'] == 'slaClass':
                        for item3 in item2['parameter']:
                            if item3["field"] == 'preferredColor':
                                item3["value"] = new_color

        # Update policy app route policy

        payload = {
            "name": policy_def["name"],
            "type": policy_def["type"],
            "description": policy_def["description"],
            "sequences": policy_def["sequences"]
        }

        api = "template/policy/definition/%s/%s" % (policy_type, policy_id)
        url = self.base_url + api
        response = HttpMethods(self.session, url).request('PUT', payload=json.dumps(payload))

        if response['status_code'] == 200:
            master_templates_affected = response['json']['masterTemplatesAffected']
            if master_templates_affected:
                print("\nMaster templates affected: %s" % master_templates_affected)
            else:
                print("\nUpdated %s App Route policy successfully" % (name))
                sys.exit()
        else:
            error = response['error']
            result = response['details']
            raise Exception(f'{error}: {result}')

        # Get device uuid and csv variables for each template id which is affected by prefix list edit operation

        inputs = []

        for template_id in master_templates_affected:
            device_ids = self.get_device_ids(template_id)
            device_inputs = self.get_device_inputs(template_id, device_ids)
            inputs.append((template_id, device_inputs))

        device_template_list = []

        for (template_id, device_input) in inputs:
            device_template_list.append({'templateId': template_id, 'isEdited': True, 'device': device_input})

        #api for CLI template 'template/device/config/attachcli'

        api = 'template/device/config/attachfeature'

        url = self.base_url + api

        payload = {'deviceTemplateList': device_template_list}

        response = HttpMethods(self.session, url).request('POST', payload=json.dumps(payload))

        if response['status_code'] == 200:
            process_id = response['json']['id']
        else:
            error = response['error']
            result = response['details']
            raise Exception(f'{error}: {result}')

        api = 'device/action/status/' + process_id

        url = self.base_url + api

        while (1):
            time.sleep(10)
            response = HttpMethods(self.session, url).request('GET')
            if response['status_code'] == 200:
                if response['json']['summary']['status'] == "done":
                    if 'Success' in response['json']['summary']['count']:
                        print("\nUpdated %s App Route policy successfully" % (name))
                    elif 'Failure' in response['json']['summary']['count']:
                        print("\nFailed to update App route policy")
                    break
            else:
                error = response['error']
                result = response['details']
                raise Exception(f'{error}: {result}')



import click
from vmanage.api.certificate import Certificate


@click.command('generate-csr')
@click.option('--ip', '-i', help="Device IP address.", required=True)
@click.option('--file', '-f', 'csr_file', help="Output file name", required=True)
@click.pass_obj
def generate_csr(ctx, ip, csr_file):
    """
    Generate CSR for a device
    """

    vmanage_certificate = Certificate(ctx.auth, ctx.host, ctx.port)
    csr = vmanage_certificate.generate_csr(ip)
    with open(csr_file, 'w') as outfile:
        outfile.write(csr)
